from collections.abc import Sequence
from typing import Self

from factrainer.base.dataset import (
    BaseDataset,
    IndexableDataset,
    RowIndex,
    RowIndices,
)
from sklearn.model_selection._split import _BaseKFold


class IndexedDataset[T: IndexableDataset](BaseDataset):
    index: RowIndex
    data: T

    def __len__(self) -> int:
        return len(self.index)


class IndexedDatasets[T: IndexableDataset](BaseDataset):
    datasets: Sequence[IndexedDataset[T]]

    def __len__(self) -> int:
        return sum([len(dataset) for dataset in self.datasets])

    @classmethod
    def create(cls, dataset: T, k_fold: _BaseKFold | RowIndices) -> Self:
        if isinstance(k_fold, _BaseKFold):
            raise NotImplementedError
        return cls(
            datasets=[
                IndexedDataset(index=index, data=dataset[index]) for index in k_fold
            ]
        )

    @property
    def indices(self) -> RowIndices:
        return [dataset.index for dataset in self.datasets]


class SplittedDataset[T: IndexableDataset](BaseDataset):
    train: IndexedDataset[T]
    val: IndexedDataset[T] | None
    test: IndexedDataset[T]


class SplittedDatasetsIndices(BaseDataset):
    train: RowIndices
    val: RowIndices | None
    test: RowIndices


class SplittedDatasets[T: IndexableDataset](BaseDataset):
    datasets: Sequence[SplittedDataset[T]]

    @property
    def train(self) -> IndexedDatasets[T]:
        return IndexedDatasets(datasets=[dataset.train for dataset in self.datasets])

    @property
    def val(self) -> IndexedDatasets[T] | None:
        vals = []
        for dataset in self.datasets:
            if dataset.val is None:
                return None
            vals.append(dataset.val)
        return IndexedDatasets(datasets=vals)

    @property
    def test(self) -> IndexedDatasets[T]:
        return IndexedDatasets(datasets=[dataset.test for dataset in self.datasets])

    @property
    def indices(self) -> SplittedDatasetsIndices:
        return SplittedDatasetsIndices(
            train=self.train.indices,
            val=self.val.indices if self.val is not None else None,
            test=self.test.indices,
        )

    @classmethod
    def create(
        cls, dataset: T, k_fold: _BaseKFold, share_holdouts: bool = True
    ) -> Self:
        datasets = []
        for train_index, val_index in dataset.get_index(k_fold):
            if share_holdouts:
                test_index = val_index
            else:
                raise NotImplementedError
            datasets.append(
                SplittedDataset(
                    train=IndexedDataset(index=train_index, data=dataset[train_index]),
                    val=IndexedDataset(index=val_index, data=dataset[val_index]),
                    test=IndexedDataset(index=test_index, data=dataset[test_index]),
                )
            )
        return cls(datasets=datasets)



from factrainer.base.config import (
    BaseMlModelConfig,
    BasePredictConfig,
    BaseTrainConfig,
)
from factrainer.base.dataset import BaseDataset, Prediction
from factrainer.base.raw_model import RawModel

from .trait import (
    PredictorTrait,
    ValidatableTrainerTrait,
)


class SingleModelContainer[
    T: BaseDataset,
    U: RawModel,
    V: BaseTrainConfig,
    W: BasePredictConfig,
](ValidatableTrainerTrait[T, V], PredictorTrait[T, U, W]):
    def __init__(
        self,
        model_config: BaseMlModelConfig[T, U, V, W],
    ) -> None:
        self.model_config = model_config

    def train(self, train_dataset: T, val_dataset: T | None = None) -> None:
        self._model = self.model_config.learner.train(
            train_dataset, val_dataset, self.model_config.train_config
        )

    def predict(self, dataset: T) -> Prediction:
        return self.model_config.predictor.predict(
            dataset, self.raw_model, self.model_config.pred_config
        )

    @property
    def raw_model(self) -> U:
        return self._model

    @property
    def train_config(self) -> V:
        return self.model_config.train_config

    @train_config.setter
    def train_config(self, config: V) -> None:
        self.model_config.train_config = config

    @property
    def pred_config(self) -> W | None:
        return self.model_config.pred_config

    @pred_config.setter
    def pred_config(self, config: W | None) -> None:
        self.model_config.pred_config = config
